"use strict";

var progressVal = 0;

var show_progress = 5;

var start_hist_delay = 1000; // used to be 4000
var mini_delay = 300;


//var hasLocalStorage = (typeof(Storage) !== "undefined"); // ****
var hasLocalStorage = false;

var quantHSLHistogramArray;

//var saveOnServer=true;
var saveOnServer=false;
var retrieveFromServer=true;


function runlength_encode(int_array) {

    var array_len = int_array.length;

    if (array_len<2) {
	console.warn("runlength_encode(): unable to run-length encode array of length " + array_len);
	console.warn("returning null");
	return null;
    }

    var int_array_rle = new Array();


    var lock_val=int_array[0];
    var next_val=int_array[1];
    var i = 1;

    while (i<array_len) {
	var c=1;
	while (next_val==lock_val) {
	    c++;
	    var next_pos = i+c-1;
	    if (next_pos<array_len) {
		next_val = int_array[next_pos];
	    }
	    else {
		// hit end of array
		break;
	    }
	}
	// come out of a run-length sequence of same val
	// next_val is one past the run-length sequence

	int_array_rle.push(c);
	int_array_rle.push(lock_val);

	i+=c;

	if (i<array_len) {
	    lock_val = next_val;
	    next_val = int_array[i];
	}
	else {
	    break;
	}
    }

    return int_array_rle;
}



function runlength_decode(int_array_rle) {

    var array_rle_len = int_array_rle.length;

    if (array_rle_len<2) {
	console.warn("runlength_decode(): unable to run-length decode array of length " + array_rle_len);
	console.warn("returning null");
	return null;
    }

    var int_array = new Array();
    for (var i=0; i<array_rle_len; i+=2){
	var rep = int_array_rle[i];
	var val = int_array_rle[i+1];
	for (var r=0; r<rep; r++) {
	    int_array.push(val);
	}
    }

    return int_array;
}



function arrays_identical(int_array1, int_array2) {

    var array_len1 = int_array1.length;
    var array_len2 = int_array2.length;

    if (array_len1 != array_len2) {
	console.warn("arrays_identical(): array lengths differ, array1("+array_len1+") vs array2(" + array_len2 + ")");
	return false;
    }

    var all_ident = true;

    for (var i=0; i<array_len1; i++){
	if (int_array1[i] != int_array2[i]) {
	    all_ident = false;
	    break;
	}
    }

    return all_ident;
}


function drawImage(imageObj, canvasId) {
    var imgCanvas = document.getElementById(canvasId),
        context, imageData, data,
        imageX = 0, imageY = 0,
        imageWidth = imageObj.width,
        imageHeight = imageObj.height;

    context = imgCanvas.getContext('2d');
    context.drawImage(imageObj, imageX, imageY, imageWidth, imageHeight);
    imageData = context.getImageData(imageX, imageY, imageWidth, imageHeight);

    return imageData;
}



function createHSLHistogramFromCanvasData(imageData, imageWidth, imageHeight, quant)
{
    var i;

    var outputLength = quant * quant * quant;
    var outputArray = new Array(outputLength);
    for (i=0; i<outputLength; i++) {
	outputArray[i] = 0;
    }

    var total = 0.0;
            
    var log2 = Math.log(2);

    //calculate value to shift by (log base 2)
	    
    var shift_sat = Math.floor(Math.log(quant)/log2);
    var shift_hue = Math.floor(Math.log(quant*quant)/log2);

    //console.log("bit-shift sat = " + shift_sat);
    //console.log("bit-shift hue = " + shift_hue);

    var data = imageData.data;

    //Loop through each pixel
    var y,p;
    for (y=0,p=0; y<imageHeight; y++) {
  	  var x;
	  for (x=0; x<imageWidth; x++,p+=4) {
	    // convert 4 byte data value into a colour pixel
	    var rgb_c = new RGBColour(data[p],data[p+1],data[p+2],data[p+3]/255.0);
	    var hsl = new rgb_c.getHSL();

	    // express h,s and b in the range 0..1

	    var h = hsl.h / 360.0;
	    var s = hsl.s / 100.0;
	    var b = hsl.l / 100.0;
	    
	    var result =    (Math.floor(h * (quant - 1)) << shift_hue) 
		          | (Math.floor(s * (quant - 1)) << shift_sat) 
		          |  Math.floor(b * (quant - 1));

	    //number of opaque pixels
	    var count = hsl.a;  // 'a' is naturally in the range 0..1
		    
	    //add to the total number of opaque pixels
	    total += count;

	    //add the frequency of each colour to the histogram
	    outputArray[result] += count;                    
	  }
    }

    if (total == 0) {
	// entire image must have been blank
	// make total '1' to stop NaN being generated
	total = 1;
    }

    //Normalise the histogram, based on the number of opaque pixels
    for (i=0; i<outputArray.length; i++) {
	outputArray[i] = outputArray[i] / total;
    }

    return outputArray;
}


function createHSLHistogram(imageObj, quant)
{
    if (!$('#hsl_canvas').length) {
	$('<canvas>', {id: 'hsl_canvas'}).hide().appendTo('body');
    }

    var imageData = drawImage(imageObj,'hsl_canvas');

    var outputArray = createHSLHistogramFromCanvasData(imageData,imageObj.width,imageObj.height,quant);

/*
    if (saveOnServer) {
	console.log("Saving HSV quantized histogram on server: " + imageObj.title.toLowerCase())
	//console.log("outputArray = " + JSON.stringify(outputArray));

	var jsonFilename = imageObj.title.toLowerCase() + ".json";
	var data = { jsonFilename: jsonFilename, jsonData: JSON.stringify(outputArray)};

	console.log("Saving colour histogram data for " + imageObj.title.toLowerCase());
	$.ajax({
  	    type: "POST",
	    url: "../similarity-2d/save-json-data.jsp",
	    data: data
	});
    }
*/

    return outputArray;
}

var flag_comparison_array = Array(2);
var clicked_on_count = 0;

function quantizedColourHistogramComparison(quant_colour_hist1,quant_colour_hist2) 
{
    // Assumes both histograms are of the same length

    var i;
    var quant_product = 0.0;

    //Loop through and compare the histograms
    for (i=0; i<quant_colour_hist1.length; i++) {
	//take the overlap
	quant_product += Math.min(quant_colour_hist1[i], quant_colour_hist2[i]);

	// consider replacing with abs for difference
    }

    console.log("Product for HSL histogram: " + quant_product);

    return quant_product;
}


function displayFlags(img_list,$displayDiv,$progressArea,$progressBar)
{
    var img_list_len = img_list.length;

    if (img_list_len==0) {
	return;
    }
    else if (img_list_len>show_progress) {
	$progressArea.slideDown();
    }

    img_list.sort();

    var root_img_re = /^.*\/(.*?)\..*?$/;

    var progress_step = 100.0 / img_list_len;
    
    var callback_count = 0;
    
    var i;
    for (i=0; i<img_list.length; i++) {
	var img_url     = img_list[i];
	var img_matches = root_img_re.exec(img_url);
	var title       = img_matches[1];
	
	var imageObj = new Image();
	imageObj.onload = function() {
	    callback_count++;
	    progressVal += progress_step;	
	    $progressBar.progressbar('value',progressVal);
	    
	    if (callback_count == img_list_len) {
		if (img_list_len>show_progress) {
		    progressVal = 100;
		    $progressBar.progressbar('value',progressVal);
		    //$progressArea.slideUp();

		    // Now move on and start computing colour histograms
		    setTimeout(function() 
			       {  doneDisplayFlags(img_list,$progressArea,$progressBar); }
			       ,start_hist_delay);
		}
	    }
	};


	/*
$("h2").live('mousedown', function(e) { 
   if( (e.which == 1) ) {
     alert("left button");
   }if( (e.which == 3) ) {
     alert("right button");
   }else if( (e.which == 2) ) {
      alert("middle button"); 
   }
   e.preventDefault();
}).live('contextmenu', function(e){
 e.preventDefault();
});
	*/

	imageObj.onclick = function(e) {
	    // Store result of HSL quantization in the image object/tag
	    if (!this.quantHSLHist) {
		// Only need to compute this if it is not already stored in the object/tag
		this.quantHSLHist = createHSLHistogram(this,16);
		console.log(this.quantHSLHist);
	    }

	    if (e.which == 1 ) {
		// Left => use in comparison pair
		flag_comparison_array[clicked_on_count] = this.quantHSLHist;

		clicked_on_count++;
		if (clicked_on_count==2) {
		    // trigger comparison
		    alert("Similarity score: " + quantizedColourHistogramComparison(flag_comparison_array[0],flag_comparison_array[1]));
		    clicked_on_count=0;
		}
	    }
	    else if (e.which == 2) {
		// Middle => fix on this and do similarity with everything else
	    	var fix_id = this.id;
		e.preventDefault();
	    }

	};
	
	
	imageObj.src   = img_url;
	imageObj.title = title.toUpperCase();
	imageObj.country = imageObj.title;
	imageObj.id    = "flag-img-" + i;

	$displayDiv.append(imageObj);
    }
}





function calcSimilarityValues(newFlagQuantHSLHist, $displayDiv,$progressArea,$progressBar)
{
    console.log("*** calcSimilarityValues()");

    var iframe = $('#similarity-2d-iframe').contents();

    var img_obj_list = iframe.find('#flagArea').children('img').toArray();

    var img_obj_list_len = img_obj_list.length;

    if (img_obj_list_len==0) {
	return;
    }
    else if (img_obj_list_len>show_progress) {
	// number of images to process non-trivial
	$progressArea.slideDown();
    }

    var progress_step = 100.0 / img_obj_list_len;
    
    for (var i=0; i<img_obj_list.length; i++) {
	var imgObj = img_obj_list[i];

	var existingQuantHSLHist = imgObj.quantHSLHist;

	var similarity_score = quantizedColourHistogramComparison(newFlagQuantHSLHist,existingQuantHSLHist);
	imgObj.similarityScore = similarity_score;
	imgObj.title = imgObj.country + ", Similarity score: " + similarity_score;

	progressVal += progress_step;	
	console.log("*** i = " + i);
	console.log("*** pbar = " + $progressBar.progressbar);
	if ($progressBar.progressbar) {
	    $progressBar.progressbar('value',progressVal);
	}
    }

    //progressVal = 100;
    //$progressBar.progressbar('value',progressVal);

    if (img_obj_list_len>show_progress) {
	$progressArea.slideUp();
    }
}

function sortUsingAttribute(parent, childSelector, field) {
    var items = parent.children(childSelector).sort(function(a, b) {
        var vA = a.attr(field);
        var vB = b.attr(field)
        return (vA < vB) ? -1 : (vA > vB) ? 1 : 0;
    });

    parent.empty(childSelecctor);
    parent.append(items);
}


function compareSimilarityValue(a,b) { 

    //var vA = a.similarityScore;
    //var vB = b.similarityScore;
    //return (vA < vB) ? -1 : (vA > vB) ? 1 : 0;
    
    return (b.similarityScore - a.similarityScore); 
}

function sortFlagsBySimilarityValue($iframe)
{
    var img_obj_list = $iframe.find('#flagArea > img').toArray();

    var sorted_img_obj_list = img_obj_list.sort(compareSimilarityValue);
    //console.log("*** sorted img len = " + sorted_img_obj_list.length);

    $iframe.find('#flagArea').empty('img');

    $iframe.find('#flagArea').append(sorted_img_obj_list);

    //sortUsingAttribute('#flagArea','img','similarityValue');
}



function computeFlagHistogramChain(chain_count, img_list_len,
				   progressVal, progress_step,
				   $progressArea,$progressBar)
{
    var flag_id = "flag-img-" + chain_count;
    console.log("Processing Image ID: " + flag_id);
    var imageObj = document.getElementById(flag_id);

    if (!imageObj.quantHSLHist) {
	// Only need to compute this if it is not
	// already stored in the object/tag

	if (retrieveFromServer) {
	    // see if it is stored on the server 
	    console.log("Retrieving stored colour histogram for " + imageObj.title.toLowerCase() + " ...");
	    $.ajax({
		async: false,
		dataType: "json",
		url: "../similarity-2d/json-data/" + imageObj.title.toLowerCase() + ".json",
		success: function(data) {
		    console.log("... success");
		    // 'data' is the array of integer vals we want
		    imageObj.quantHSLHist = data;
		},
		error: function() {
		    console.log("... not stored on server");
		    // not stored on the server (currently)
		    // => need to compute it
		    imageObj.quantHSLHist = createHSLHistogram(imageObj,16);
		    //console.log(imageObj.quantHSLHist);

		}
	    });
	}
	else {
	    imageObj.quantHSLHist = createHSLHistogram(imageObj,16);
	}

    }

    if (chain_count < img_list_len-1) {
	progressVal += progress_step;	
	$progressBar.progressbar('value',progressVal);

	setTimeout(function()
		   { computeFlagHistogramChain(chain_count+1,img_list_len,
					       progressVal, progress_step,
					       $progressArea,$progressBar)
			   }
		   ,mini_delay // any time delay causes break in 'rendering' thread, even 0!
		   );

    }
    else {
	$progressArea.find('span:first').text("Done.");
	if (img_list_len>show_progress) {
	    progressVal = 100;
	    $progressBar.progressbar('value',progressVal);
	    $progressArea.slideUp();
	}

	doneComputingHistograms(img_list_len);
    }
}


function computeFlagHistograms(img_list,$progressArea,$progressBar)
{
    var img_list_len = img_list.length;


    if (img_list_len==0) {
	$progressArea.find('span:first').text("Empty list: no computation needed.");
	return;
    }


    $progressArea.find('span:first').text("Computing histograms:");
    progressVal = 0;

    if (img_list_len>show_progress) {
	$progressArea.slideDown();
    }

    var progress_step = 100.0 / img_list_len;
 
    // Start the chaining process
    setTimeout(function() 
	       { computeFlagHistogramChain(0,img_list_len,
					   progressVal, progress_step,
					   $progressArea,$progressBar)
		       }
	       ,mini_delay // any time delay causes break in 'rendering' thread, even 0!
	       );
    
}

function doneComputingHistograms(img_list_len)
{
    quantHSLHistogramArray = new Array(img_list_len);
    for (var i=0; i<img_list_len; i++) {
	var flag_id = "flag-img-" + i;
	var imageObj = document.getElementById(flag_id);
	quantHSLHistogramArray[i] = imageObj.quantHSLHist;
    } 
    var json_hist_data = JSON.stringify(quantHSLHistogramArray);

    if (hasLocalStorage) {
	// build up quantHSLHistogramArray from all the images

	console.log("Storing computed HSL histograms in localStorage");
	localStorage.quantHSLHistogramArray = json_hist_data;
    }

    if (saveOnServer) {
	// run-length encode data, so less to send
	var quantHSLHistogramArrayRLE = new Array(img_list_len);
	for (var i=0; i<img_list_len; i++) {
	    //console.log("rle encoding ... i = " + i);
	    quantHSLHistogramArrayRLE[i] = runlength_encode(quantHSLHistogramArray[i])

	    //var test_decode_array = runlength_decode(quantHSLHistogramArrayRLE[i]);
	    //console.log("Checking decoded array: " + arrays_identical(test_decode_array,quantHSLHistogramArray[i]));
	}


	var json_hist_data_rle = JSON.stringify(quantHSLHistogramArrayRLE);

	var json_filename = "all-hsv-histograms-rle.json";
	var data = { jsonFilename: json_filename, jsonData: json_hist_data_rle};
	
	console.log("Saving colour histogram data for all flags");
	$.ajax({
  	    type: "POST",
	    url: "../similarity-2d/save-json-data.jsp",
	    data: data	    
	});
    }
}

function doneDisplayFlags(img_list,$progressArea,$progressBar)
{
    console.log("doneDisplayFlags()");

    var needToComputeHistograms = true;

    if (hasLocalStorage) {

	if (localStorage.quantHSLHistogramArray) {

	    console.log("Retrieving quantized HSL histograms from local storage");

	    var stored_hist_array = JSON.parse(localStorage["quantHSLHistogramArray"]);
	    if (stored_hist_array.length == img_list.length) {
		// Assume if the number of images process hasn't changed, then neither
		// has the stored historgram values

		quantHSLHistogramArray = stored_hist_array;

		var i;
		for (i=0; i<img_list.length; i++) {
		    var flag_id = "flag-img-" + i;
		    var imageObj = document.getElementById(flag_id);
		    imageObj.quantHSLHist = quantHSLHistogramArray[i];
		} 

		needToComputeHistograms = false;
	    }
	    else {
		console.log("Different number of images detected => regenerating");
	    }
	}
    }

    if (needToComputeHistograms) {

	if (retrieveFromServer) {
	    // before computing histograms from scratch, see if resulting data stored on server

	    console.log("Retrieving stored colour histgram data for all flags ...");
	    $.ajax({
		async: true,
		dataType: "json",
		url: "../similarity-2d/json-data/all-hsv-histograms-rle.json",
		success: function(rle_data) {
		    console.log("... success");
		    if (rle_data.length != img_list.length) {
			console.log("Stored histogram data out of date with flag list.");
			console.log("rle len = " + rle_data.length + ", img_list len = " + img_list.length);
			// what is stored on the sever differed to the number of images displayed
			computeFlagHistograms(img_list,$progressArea,$progressBar);
		    }
		    else {
			var img_list_len = img_list.length;
			quantHSLHistogramArray = new Array(img_list_len);
			
			// Need to rle decode each entry in this array
			for (var i=0; i<img_list_len; i++) {
			    quantHSLHistogramArray[i] = runlength_decode(rle_data[i]);
			    var flag_id = "flag-img-" + i;
			    var imageObj = document.getElementById(flag_id);
			    imageObj.quantHSLHist = quantHSLHistogramArray[i];

			}

			needToComputeHistograms = false;

			$progressArea.slideUp();
			doneComputingHistograms(img_list.length);

		    }
		},
		error: function() {
		    console.log("... not stored on server");
		    computeFlagHistograms(img_list,$progressArea,$progressBar);
		}
	    });


	}
	else {
	    computeFlagHistograms(img_list,$progressArea,$progressBar);
	}
    }
    else {
	$progressArea.slideUp();
	doneComputingHistograms(img_list.length);
    }
}

function rankBySimilarity(imageData,x_dim,y_dim)
{
    // quantizedColourHistogramComparison(quant_colour_hist1,quant_colour_hist2) 

}
